/*
 *  Copyright (C) 2006-10 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "nprobe.h"

#define BASE_ID            NTOP_BASE_ID+180
#define URL_MAX_LEN        64
#define HOST_MAX_LEN       32

#define MAX_BYTES_SENT     64
#define MAX_BYTES_RCVD     64

/* RFC 2616 */
/* Request */
#define GET_URL            "GET /"
#define POST_URL           "POST /"
#define HEAD_URL           "HEAD /"
/* Note the other methods (PUT, DELETE, TRACE, CONNECT)
   have not been implemented as they are used very seldom */
#define HOST_STR           "Host: "
#define REFERER_STR        "Referer: "
#define USER_AGENT_STR     "User-Agent: "
#define CONTENT_TYPE_STR   "Content-type: "

/* Response */
#define HTTP_1_0_URL       "HTTP/1.0 "
#define HTTP_1_1_URL       "HTTP/1.1 "

static V9V10TemplateElementId httpPlugin_template[] = {
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID,   VARIABLE_FIELD_LEN, URL_MAX_LEN, ascii_format, dump_as_ascii, "HTTP_URL", "HTTP URL" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+1, STATIC_FIELD_LEN, 2, numeric_format, dump_as_uint, "HTTP_RET_CODE", "HTTP return code (e.g. 200, 304...)" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+2, VARIABLE_FIELD_LEN, URL_MAX_LEN, ascii_format, dump_as_ascii, "HTTP_REFERER", "HTTP Referer" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+3, VARIABLE_FIELD_LEN, URL_MAX_LEN, ascii_format, dump_as_ascii, "HTTP_UA", "HTTP User Agent" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+4, VARIABLE_FIELD_LEN, URL_MAX_LEN, ascii_format, dump_as_ascii, "HTTP_MIME", "HTTP Mime Type" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, 0, STATIC_FIELD_LEN, 0, 0, 0, NULL, NULL }
};

struct plugin_info {
  char *http_url, *http_host, *referer, *user_agent, *method, *content_type;
  u_int16_t ret_code;
};

// #define DEBUG

/* *********************************************** */

static PluginInfo httpPlugin; /* Forward */

/* ******************************************* */

void httpPlugin_init(int argc, char *argv[]) {
  traceEvent(TRACE_INFO, "Initialized HTTP plugin");
}

/* *********************************************** */

/* Handler called whenever an incoming packet is received */

static void httpPlugin_packet(u_char new_bucket, void *pluginData,
			      FlowHashBucket* bkt,
			      u_short proto, u_char isFragment,
			      u_short numPkts, u_char tos,
			      u_short vlanId, struct eth_header *ehdr,
			      IpAddress *src, u_short sport,
			      IpAddress *dst, u_short dport,
			      u_int len, u_int8_t flags, u_int8_t icmpType,
			      u_short numMplsLabels,
			      u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN],
			      char *fingerprint,
			      const struct pcap_pkthdr *h, const u_char *p,
			      u_char *payload, int payloadLen) {
  PluginInformation *info;
  struct plugin_info *pinfo;

  // traceEvent(TRACE_INFO, "httpPlugin_packet(%d)", payloadLen);

  if(new_bucket) {
    info = (PluginInformation*)malloc(sizeof(PluginInformation));
    if(info == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      return; /* Not enough memory */
    }

    info->pluginPtr  = (void*)&httpPlugin;
    pluginData = info->pluginData = malloc(sizeof(struct plugin_info));

    if(info->pluginData == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      free(info);
      return; /* Not enough memory */
    } else
      memset(info->pluginData, 0, sizeof(struct plugin_info));

    info->next = bkt->plugin;
    bkt->plugin = info;
  }

  /*
    Do not put the checks below at the beginning of the function as otherwise
    the flow will not be able to export info about the template specified
   */
  if(proto != IPPROTO_TCP) return;

  if((sport != 80) && (sport != 3128) && (sport != 8080)
     && (dport != 80) && (dport != 3128) && (dport != 8080))
    return;

  pinfo = (struct plugin_info*)pluginData;

  if(payloadLen > 0) {
    char *method;
    u_short shift;

    //traceEvent(TRACE_INFO, "==> [%d][%d]'%s'", bkt->bytesSent, bkt->bytesRcvd, payload);

    payload[payloadLen] = '\0';

    if((!strncmp((char*)payload, GET_URL, strlen(GET_URL)))) method = GET_URL, shift = 1;
    else if((!strncmp((char*)payload, POST_URL, strlen(POST_URL)))) method = POST_URL, shift = 1;
    else if((!strncmp((char*)payload, HEAD_URL, strlen(HEAD_URL)))) method = HEAD_URL, shift = 1;
    else if((!strncmp((char*)payload, HTTP_1_0_URL, strlen(HTTP_1_0_URL)))) method = HTTP_1_0_URL, shift = 0;
    else if((!strncmp((char*)payload, HTTP_1_1_URL, strlen(HTTP_1_1_URL)))) method = HTTP_1_1_URL, shift = 0;
    else method = NULL;

    if(method) {
      char url[1024] = { 0 }, *referer, *host, *user_agent, *content_type;
      int i;

      len = strlen(method)-shift;
      strncpy(url, (char*)&payload[len], min(sizeof(url), (payloadLen-len)));

      url[sizeof(url)-1] = '\0';
      for(i=0; i<sizeof(url); i++)
	if((url[i] == ' ')
	   || (url[i] == '\r')
	   || (url[i] == '\n')) {
	  url[i] = '\0';
	  break;
	}

      if((!strncmp(method, GET_URL, strlen(GET_URL)))
	 || (!strncmp(method, POST_URL, strlen(POST_URL)))
	 || (!strncmp(method, HEAD_URL, strlen(HEAD_URL)))
	 ) {
	/* HTTP Request */

	/* We need to export this flow now */
	if(pinfo->http_url != NULL) {
	  exportBucket(bkt, 0);
	  resetBucketStats(bkt, h, len, sport, dport, payload, payloadLen);
	  free(pinfo->http_url);
	  pinfo->http_url = NULL;

	  if(pinfo->referer != NULL)    { free(pinfo->referer); pinfo->referer = NULL; }
	  if(pinfo->user_agent != NULL) { free(pinfo->user_agent); pinfo->user_agent = NULL; }
	  if(pinfo->content_type != NULL) { free(pinfo->content_type); pinfo->content_type = NULL; }
	}

	pinfo->method = method;

	host = strstr((char*)payload, HOST_STR);
	if(host != NULL) {
	  char *str = &host[sizeof(HOST_STR)-1], old_char = 0;

	  if(pinfo->http_host != NULL) free(pinfo->http_host);

	  for(i=0; str[i] != '\0'; i++) {
	    if((str[i] == ' ')
	       || (str[i] == '\r')
	       || (str[i] == '\n')) {
	      old_char = str[i];
	      str[i] = '\0';
	      break;
	    }
	  }

	  pinfo->http_host = strdup(str);
	  if(old_char != 0) str[i] = old_char;
#ifdef DEBUG
	  if(pinfo->http_host) traceEvent(TRACE_INFO, "==> HOST='%s'", pinfo->http_host);
#endif
	}

	referer = strstr((char*)payload, REFERER_STR);
	if(referer != NULL) {
	  char *str = &referer[sizeof(REFERER_STR)-1], old_char = 0;

	  /* Skip http://.... */
	  if(strncmp(str, "http://", 7) == 0) str = &str[7];

	  if(pinfo->referer != NULL) free(pinfo->referer);

	  for(i=0; str[i] != '\0'; i++) {
	    if((str[i] == ' ')
	       || (str[i] == '\r')
	       || (str[i] == '\n')) {
	      old_char = str[i];
	      str[i] = '\0';
	      break;
	    }
	  }

	  pinfo->referer = strdup(str);
	  if(old_char != 0) str[i] = old_char;

#ifdef DEBUG
	  if(pinfo->referer) traceEvent(TRACE_INFO, "==> REFERER='%s'", pinfo->referer);
#endif
	}

	user_agent = strstr((char*)payload, USER_AGENT_STR);
	if(user_agent != NULL) {
	  char *str = &user_agent[sizeof(USER_AGENT_STR)-1], old_char = 0;

	  if(pinfo->user_agent != NULL) free(pinfo->user_agent);

	  for(i=0; str[i] != '\0'; i++) {
	    if((str[i] == '\r') || (str[i] == '\n')) {
	      old_char = str[i];
	      str[i] = '\0';
	      break;
	    }
	  }

	  pinfo->user_agent = strdup(str);
	  if(old_char != 0) str[i] = old_char;

#ifdef DEBUG
	  if(pinfo->user_agent) traceEvent(TRACE_ERROR, "==> USER_AGENT='%s'", pinfo->user_agent);
#endif
	}

	if(pinfo->http_url) { free(pinfo->http_url); pinfo->http_url = NULL; }
	pinfo->http_url = strdup(url);
      } else {
	/* HTTP Response */
	pinfo->ret_code = atoi(url);

	content_type = strcasestr((char*)payload, CONTENT_TYPE_STR);
	if(content_type != NULL) {
	  char *str = &content_type[sizeof(CONTENT_TYPE_STR)-1], old_char = 0;

	  if(pinfo->content_type != NULL) free(pinfo->content_type);

	  for(i=0; str[i] != '\0'; i++) {
	    if((str[i] == '\r') || (str[i] == '\n') || (str[i] == ';')) {
	      old_char = str[i];
	      str[i] = '\0';
	      break;
	    }
	  }

	  pinfo->content_type = strdup(str);
	  if(old_char != 0) str[i] = old_char;

#ifdef DEBUG
	  if(pinfo->content_type) traceEvent(TRACE_ERROR, "==> CONTENT_TYPE='%s'", pinfo->content_type);
#endif
	}
      }
    }
  }
}

/* *********************************************** */

static char *method2str(char* method) {
  if(method == (char*)GET_URL)       return("GET");
  else if(method == (char*)POST_URL) return("POST");
  else if(method == (char*)HEAD_URL) return("HEAD");
  else return("");
}

/* *********************************************** */

/* Handler called when the flow is deleted (after export) */

#define MAX_NUM_FILES      10000 
#define MAX_NUM_LINES      10000 

static FILE *fd = NULL;
static u_int num_file = 0, num_log_lines = 0;
static u_char dump_logs = 0;

static void httpPlugin_delete(FlowHashBucket* bkt, void *pluginData) {
  if(pluginData != NULL) {
    struct plugin_info *pinfo = (struct plugin_info*)pluginData;

    if(pinfo->http_url) {
#if 0
      char buf[64], when[64], *host;
      struct tm tm;
      time_t when_time;

      host = _intoa((bkt->dport == 80) ? bkt->src->host : bkt->dst->host, buf, sizeof(buf));
      when_time = (bkt->dport == 80) ? bkt->flowTimers.firstSeenSent.tv_sec : bkt->flowTimers.firstSeenRcvd.tv_sec;
      /* 80.8.55.1 - - [01/Jan/2001:00:00:10 +0100] "GET /page1.html HTTP/1.0" 200 7009 "-" "Mozilla/4.0 (compatible; MSIE 5.5; Windows NT 5.0)" */
      localtime_r(&when_time, &tm);

      strftime(when, sizeof(when), "%d/%B/%Y:%H:%M:%S %z", &tm);

      printf("%s - - [%s] \"%s %s\" %d %u \"-\" \"%s\" \"%s\"\n",
	     host, when, method2str(pinfo->method), pinfo->http_url, pinfo->ret_code,
	     (bkt->flowCounters.bytesSent + bkt->flowCounters.bytesRcvd),
	     pinfo->referer ? pinfo->referer : "",
	     pinfo->user_agent ? pinfo->user_agent : "");
#else
      if(dump_logs) {
	char buf[64];
	char *client = _intoa((bkt->dport == 80) ? bkt->src->host : bkt->dst->host, buf, sizeof(buf));
	u_int begin_time = (bkt->dport == 80) ? bkt->flowTimers.firstSeenSent.tv_sec : bkt->flowTimers.firstSeenRcvd.tv_sec;
	u_int end_time = (bkt->dport == 80) ? bkt->flowTimers.lastSeenSent.tv_sec : bkt->flowTimers.lastSeenRcvd.tv_sec;

	if(pinfo->http_host == NULL) {
	  char buf1[64];

	  pinfo->http_url = strdup(_intoa((bkt->sport == 80) ? bkt->src->host : bkt->dst->host, buf1, sizeof(buf1))); 
	}

	/* 192.168.0.200|www.macintouch.com|GET|/images/filewave02.gif|200|www.macintouch.com/|(null)||25505|1133966827|1133966829 */

	if(fd == NULL) {
	  char buf[256];

	  snprintf(buf, sizeof(buf), "/tmp/http_%d.txt", num_file++);
	  num_file = num_file % MAX_NUM_FILES;
	  num_log_lines = 0;
	  fd = fopen(buf, "w");
	}

	if(fd) {
	  fprintf(fd, "%s|%s|%s|%s|%u|"
		  "%s|%s|%s|"
		  "%u|%u|%u\n",
		  client, pinfo->http_host, method2str(pinfo->method), pinfo->http_url, pinfo->ret_code,
		  pinfo->referer ? pinfo->referer : "", pinfo->user_agent ? pinfo->user_agent : "", pinfo->content_type ? pinfo->content_type : "",
		  (bkt->flowCounters.bytesSent + bkt->flowCounters.bytesRcvd), begin_time, end_time);
	  num_log_lines++;

	  if(num_log_lines == MAX_NUM_LINES) {
	    fclose(fd);
	    fd = NULL;
	  }
	}
      }
#endif
    }

    if(pinfo->http_url)   free(pinfo->http_url);
    if(pinfo->http_host)  free(pinfo->http_host);
    if(pinfo->referer)    free(pinfo->referer);
    if(pinfo->user_agent) free(pinfo->user_agent);
    if(pinfo->content_type) free(pinfo->content_type);

    free(pluginData);
  }
}

/* *********************************************** */

/* Handler called at startup when the template is read */

static V9V10TemplateElementId* httpPlugin_get_template(char* template_name) {
  int i;

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(!strcmp(template_name, httpPlugin_template[i].templateElementName)) {
      return(&httpPlugin_template[i]);
    }
  }

  return(NULL); /* Unknown */
}

/* *********************************************** */

static int get_url(struct plugin_info *info, char *url_buf,
		   uint urlbuf_len, u_int16_t compressData) {
  return(snprintf(url_buf, urlbuf_len, "%s%s",
		  info->http_host ? info->http_host : "",
		  info->http_url ? info->http_url : "" ));
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be exported */

static int httpPlugin_export(void *pluginData, V9V10TemplateElementId *theTemplate,
			     int direction /* 0 = src->dst, 1 = dst->src */,
			     FlowHashBucket *bkt, char *outBuffer,
			     uint* outBufferBegin, uint* outBufferMax) {
  int i;

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == httpPlugin_template[i].templateElementId) {
      if((*outBufferBegin)+httpPlugin_template[i].templateElementLen > (*outBufferMax))
	return(-2); /* Too long */

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	char url_buf[512] = { 0 };
	uint len;

	switch(httpPlugin_template[i].templateElementId) {
	case BASE_ID:
	  get_url(info, url_buf, sizeof(url_buf)-1, 1);

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (httpPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(url_buf), httpPlugin_template[i].templateElementLen);

	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }
	  } else
	    len = httpPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], url_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> URL='%s'", url_buf);
#endif
	  (*outBufferBegin) += len;
	  break;

	case BASE_ID+1:
	  copyInt16(info->ret_code, outBuffer, outBufferBegin, outBufferMax);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> RetCode='%d'", info->ret_code);
#endif
	  break;

	case BASE_ID+2:
	  if(info->referer == NULL)
	    url_buf[0] = '\0';
	  else
	    snprintf(url_buf, sizeof(url_buf)-1, "%s", info->referer);

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (httpPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(url_buf), httpPlugin_template[i].templateElementLen);

	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }
	  } else
	    len = httpPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], url_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> REFERER='%s'", url_buf);
#endif
	  (*outBufferBegin) += len;
	  break;

	case BASE_ID+3:
	  if(info->user_agent == NULL)
	    url_buf[0] = '\0';
	  else
	    snprintf(url_buf, sizeof(url_buf)-1, "%s", info->user_agent);

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (httpPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(url_buf), httpPlugin_template[i].templateElementLen);

	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }
	  } else
	    len = httpPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], url_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> USER_AGENT='%s'", url_buf);
#endif
	  (*outBufferBegin) += len;
	  break;

	case BASE_ID+4:
	  if(info->content_type == NULL)
	    url_buf[0] = '\0';
	  else
	    snprintf(url_buf, sizeof(url_buf)-1, "%s", info->content_type);

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (httpPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(url_buf), httpPlugin_template[i].templateElementLen);

	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }
	  } else
	    len = httpPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], url_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> CONTENT_TYPE='%s'", url_buf);
#endif
	  (*outBufferBegin) += len;
	  break;

	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be printed on file */

static int httpPlugin_print(void *pluginData, V9V10TemplateElementId *theTemplate,
			    int direction /* 0 = src->dst, 1 = dst->src */,
			    FlowHashBucket *bkt, char *line_buffer, uint line_buffer_len) {
  int i;
  char url_buf[256] = { 0 };

  for(i=0; httpPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == httpPlugin_template[i].templateElementId) {

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;

	switch(httpPlugin_template[i].templateElementId) {
	case BASE_ID:
	  get_url(info, url_buf, sizeof(url_buf), 0);
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", url_buf);
	  break;
	case BASE_ID+1:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->ret_code);
	  break;
	case BASE_ID+2:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", info->referer ? info->referer : "");
	  break;
	case BASE_ID+3:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", info->user_agent ? info->user_agent : "");
	  break;
	case BASE_ID+4:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", info->content_type ? info->content_type : "");
	  break;
	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

static V9V10TemplateElementId* httpPlugin_conf(void) {
  return(httpPlugin_template);
}

/* *********************************************** */

/* Plugin entrypoint */
static PluginInfo httpPlugin = {
  NPROBE_REVISION,
  "HTTP Protocol Dissector",
  "0.1",
  "Handle HTTP protocol",
  "L.Deri <deri@ntop.org>",
  0 /* not always enabled */, 1, /* enabled */
  httpPlugin_init,
  NULL, /* Term */
  httpPlugin_conf,
  httpPlugin_delete,
  1, /* call packetFlowFctn for each packet */
  httpPlugin_packet,
  httpPlugin_get_template,
  httpPlugin_export,
  httpPlugin_print,
  NULL,
  NULL
};

/* *********************************************** */

/* Plugin entry fctn */
#ifdef MAKE_STATIC_PLUGINS
PluginInfo* httpPluginEntryFctn(void)
#else
  PluginInfo* PluginEntryFctn(void)
#endif
{
  return(&httpPlugin);
}

