/*
 *  Copyright (C) 20010 Luca Deri <deri@ntop.org>
 *
 *  		       http://www.ntop.org/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "nprobe.h"

#define BASE_ID            NTOP_BASE_ID+195
#define SQL_MAX_LEN        32

static V9V10TemplateElementId mysqlPlugin_template[] = {
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID,   VARIABLE_FIELD_LEN, SQL_MAX_LEN, ascii_format, dump_as_ascii, "MYSQL_SQL", "MYSQL SQL" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, BASE_ID+1, STATIC_FIELD_LEN, 2, numeric_format, dump_as_uint, "MYSQL_RET_CODE", "MYSQL return code (e.g. 200, 304...)" },
  { FLOW_TEMPLATE, NTOP_ENTERPRISE_ID, 0, STATIC_FIELD_LEN, 0, 0, 0, NULL, NULL }
};

/* http://forge.mysql.com/wiki/MySQL_Internals_ClientServer_Protocol */

struct mysql_header {
  u_int8_t length[3], pkt_number;
};

struct plugin_info {
  char *mysql_sql;
  u_int16_t ret_code;
};

// #define DEBUG 

/* *********************************************** */

static PluginInfo mysqlPlugin; /* Forward */

/* ******************************************* */

void mysqlPlugin_init(int argc, char *argv[]) {
  traceEvent(TRACE_INFO, "Initialized MySQL plugin");
}

/* *********************************************** */

/* Handler called whenever an incoming packet is received */

static void mysqlPlugin_packet(u_char new_bucket, void *pluginData,
			       FlowHashBucket* bkt,
			       u_short proto, u_char isFragment,
			       u_short numPkts, u_char tos,
			       u_short vlanId, struct eth_header *ehdr,
			       IpAddress *src, u_short sport,
			       IpAddress *dst, u_short dport,
			       u_int len, u_int8_t flags, u_int8_t icmpType,
			       u_short numMplsLabels,
			       u_char mplsLabels[MAX_NUM_MPLS_LABELS][MPLS_LABEL_LEN],
			       char *fingerprint,
			       const struct pcap_pkthdr *h, const u_char *p,
			       u_char *payload, int payloadLen) {
  PluginInformation *info;
  struct plugin_info *pinfo;

  if(proto != IPPROTO_TCP) return;

  if((sport != 3306) && (dport != 3306))
    return;

  // traceEvent(TRACE_INFO, "mysqlPlugin_packet(%d)", payloadLen);

  if(new_bucket) {
    info = (PluginInformation*)malloc(sizeof(PluginInformation));
    if(info == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      return; /* Not enough memory */
    }

    info->pluginPtr  = (void*)&mysqlPlugin;
    pluginData = info->pluginData = malloc(sizeof(struct plugin_info));

    if(info->pluginData == NULL) {
      traceEvent(TRACE_ERROR, "Not enough memory?");
      free(info);
      return; /* Not enough memory */
    } else
      memset(info->pluginData, 0, sizeof(struct plugin_info));

    info->next = bkt->plugin;
    bkt->plugin = info;
  }

  pinfo = (struct plugin_info*)pluginData;

  if(payloadLen > 0) {
    /* Add code for filling up pinfo */
    traceEvent(TRACE_ERROR, "--> Parse packet [len=%d]", payloadLen);
    pinfo->ret_code = 33;
  }  
}

/* *********************************************** */

/* Handler called when the flow is deleted (after export) */

static void mysqlPlugin_delete(FlowHashBucket* bkt, void *pluginData) {
  if(pluginData != NULL) {
    struct plugin_info *pinfo = (struct plugin_info*)pluginData;
    
    if(pinfo->mysql_sql)  free(pinfo->mysql_sql);

    free(pluginData);
  }
}

/* *********************************************** */
   
/* Handler called at startup when the template is read */

static V9V10TemplateElementId* mysqlPlugin_get_template(char* template_name) {
  int i;

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++) {
    if(!strcmp(template_name, mysqlPlugin_template[i].templateElementName)) {
      return(&mysqlPlugin_template[i]);
    }
  }

  return(NULL); /* Unknown */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be exported */

static int mysqlPlugin_export(void *pluginData, V9V10TemplateElementId *theTemplate,
			     int direction /* 0 = src->dst, 1 = dst->src */,
			     FlowHashBucket *bkt, char *outBuffer,
			     uint* outBufferBegin, uint* outBufferMax) {
  int i;

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == mysqlPlugin_template[i].templateElementId) {
      if((*outBufferBegin)+mysqlPlugin_template[i].templateElementLen > (*outBufferMax))
	return(-2); /* Too long */

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	char sql_buf[512] = { 0 };
	uint len;

	switch(mysqlPlugin_template[i].templateElementId) {
	case BASE_ID:
	  snprintf(sql_buf, sizeof(sql_buf), "%s", info->mysql_sql ? info->mysql_sql : "");

	  if((readOnlyGlobals.netFlowVersion == 10)
	     && (mysqlPlugin_template[i].variableFieldLength == VARIABLE_FIELD_LEN)) {
	    len = min(strlen(sql_buf), mysqlPlugin_template[i].templateElementLen);
	    
	    if(len < 255)
	      copyInt8(len, outBuffer, outBufferBegin, outBufferMax);
	    else {
	      copyInt8(255, outBuffer, outBufferBegin, outBufferMax);
	      copyInt16(len, outBuffer, outBufferBegin, outBufferMax);
	    }    
	  } else
	    len = mysqlPlugin_template[i].templateElementLen;

	  memcpy(&outBuffer[*outBufferBegin], sql_buf, len);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> SQL='%s'", sql_buf);
#endif
	  (*outBufferBegin) += len;
	  break;
	case BASE_ID+1:
	  copyInt16(info->ret_code, outBuffer, outBufferBegin, outBufferMax);
#ifdef DEBUG
	  traceEvent(TRACE_INFO, "==> RetCode='%d'", info->ret_code);
#endif
	  break;
	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

/* Handler called whenever a flow attribute needs to be printed on file */

static int mysqlPlugin_print(void *pluginData, V9V10TemplateElementId *theTemplate,
			    int direction /* 0 = src->dst, 1 = dst->src */,
			    FlowHashBucket *bkt, char *line_buffer, uint line_buffer_len) {
  int i;

  for(i=0; mysqlPlugin_template[i].templateElementId != 0; i++) {
    if(theTemplate->templateElementId == mysqlPlugin_template[i].templateElementId) {

      if(pluginData) {
	struct plugin_info *info = (struct plugin_info *)pluginData;
	
	switch(mysqlPlugin_template[i].templateElementId) {
	case BASE_ID:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%s", 
		   info->mysql_sql ? info->mysql_sql : "");
	  break;
	case BASE_ID+1:
	  snprintf(&line_buffer[strlen(line_buffer)], (line_buffer_len-strlen(line_buffer)), "%d", info->ret_code);
	  break;
	default:
	  return(-1); /* Not handled */
	}

	return(0);
      }
    }
  }

  return(-1); /* Not handled */
}

/* *********************************************** */

static V9V10TemplateElementId* mysqlPlugin_conf(void) {
  return(mysqlPlugin_template);
}

/* *********************************************** */

/* Plugin entrypoint */
static PluginInfo mysqlPlugin = {
  NPROBE_REVISION,
  "MYSQL Protocol Dissector",
  "0.1",
  "Handle MYSQL protocol",
  "L.Deri <deri@ntop.org>",
  0 /* not always enabled */, 1, /* enabled */
  mysqlPlugin_init,
  NULL, /* Term */
  mysqlPlugin_conf,
  mysqlPlugin_delete,
  1, /* call packetFlowFctn for each packet */
  mysqlPlugin_packet,
  mysqlPlugin_get_template,
  mysqlPlugin_export,
  mysqlPlugin_print,
  NULL,
  NULL
};

/* *********************************************** */

/* Plugin entry fctn */
#ifdef MAKE_STATIC_PLUGINS
PluginInfo* mysqlPluginEntryFctn(void)
#else
  PluginInfo* PluginEntryFctn(void)
#endif
{
  return(&mysqlPlugin);
}
 
